package pl.allegro.tech.hermes.integration.helper;

import com.jayway.awaitility.Duration;
import com.mongodb.BasicDBObject;
import com.mongodb.DBCollection;
import kafka.api.ConsumerMetadataRequest;
import kafka.common.ErrorMapping;
import kafka.javaapi.ConsumerMetadataResponse;
import kafka.network.BlockingChannel;
import org.apache.curator.framework.CuratorFramework;
import pl.allegro.tech.hermes.api.PublishedMessageTraceStatus;
import pl.allegro.tech.hermes.api.SentMessageTraceStatus;
import pl.allegro.tech.hermes.api.Subscription;
import pl.allegro.tech.hermes.api.TopicName;
import pl.allegro.tech.hermes.common.config.Configs;
import pl.allegro.tech.hermes.infrastructure.zookeeper.ZookeeperPaths;
import pl.allegro.tech.hermes.test.helper.endpoint.HermesEndpoints;

import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.function.Supplier;

import static com.jayway.awaitility.Awaitility.waitAtMost;
import static pl.allegro.tech.hermes.test.helper.endpoint.TimeoutAdjuster.adjust;

public class Waiter extends pl.allegro.tech.hermes.test.helper.endpoint.Waiter {

    private final HermesEndpoints endpoints;

    private final CuratorFramework zookeeper;

    private final CuratorFramework kafkaZookeeper;

    private final ZookeeperPaths zookeeperPaths = new ZookeeperPaths(Configs.ZOOKEEPER_ROOT.getDefaultValue().toString());

    public Waiter(HermesEndpoints endpoints, CuratorFramework zookeeper, CuratorFramework kafkaZookeeper) {
        super(endpoints);
        this.endpoints = endpoints;
        this.zookeeper = zookeeper;
        this.kafkaZookeeper = kafkaZookeeper;
    }

    public void untilKafkaZookeeperNodeDeletion(final String path) {
        untilZookeeperNodeDeletion(path, kafkaZookeeper);
    }

    public void untilZookeeperNodeCreation(final String path) {
        untilZookeeperNodeCreation(path, zookeeper);
    }

    public void untilKafkaZookeeperNodeEmptied(final String path, int seconds) {
        untilZookeeperNodeEmptied(path, seconds, kafkaZookeeper);
    }

    public void untilTopicDetailsAreCreated(TopicName topicName) {
        untilZookeeperNodeCreation(zookeeperPaths.topicPath(topicName));
    }

    public void untilSubscriptionMetricsIsCreated(TopicName topicName, String subscriptionName) {
        untilZookeeperNodeCreation(zookeeperPaths.subscriptionMetricsPath(topicName, subscriptionName));
    }

    public void untilSubscriptionMetricsIsRemoved(TopicName topicName, String subscriptionName) {
        untilZookeeperNodeDeletion(zookeeperPaths.subscriptionMetricsPath(topicName, subscriptionName), zookeeper);
    }

    public void untilSubscriptionIsDeactivated(String group, String topic, String subscription) {
        untilKafkaZookeeperNodeEmptied(subscriptionConsumerPath(group, topic, subscription), 60);
    }

    public void untilSubscriptionIsActivated(String group, String topic, String subscription) {
        waitAtMost(adjust(Duration.ONE_MINUTE)).until(() -> {
            endpoints.subscription().get(group + "." + topic, subscription).getState().equals(Subscription.State.ACTIVE);
        });
    }

    public void untilSubscriptionEndsReiteration(TopicName topicName, String subscription) {
        untilSubscriptionEndsReiteration(topicName.getGroupName(), topicName.getName(), subscription);
    }

    public void untilSubscriptionEndsReiteration(final String group, final String topic, final String subscription) {
        waitAtMost(adjust(30), TimeUnit.SECONDS).until(() -> {
            Subscription.State state = endpoints.subscription().get(group + "." + topic, subscription).getState();
            return state == Subscription.State.ACTIVE;
        });
    }

    public void untilAllOffsetsEqual(final String group, final String topic, final String subscription, final int offset) {
        waitAtMost(adjust(30), TimeUnit.SECONDS).until(() -> {
            List<String> partitions = zookeeper.getChildren().forPath(subscriptionOffsetPath(group, topic, subscription));
            for (String partition : partitions) {
                Long currentOffset = Long.valueOf(new String(
                        zookeeper.getData().forPath(subscriptionOffsetPath(group, topic, subscription) + "/" + partition)));
                if (currentOffset != offset) {
                    return false;
                }
            }
            return true;
        });
    }

    public void untilConsumersRebalance(final String group, final String topic, final String subscription, final int consumerCount) {
        waitAtMost(adjust(Duration.ONE_MINUTE)).until(() -> {
            List<String> children = zookeeper.getChildren().forPath(subscriptionIdsPath(group, topic, subscription));
            return children != null && children.size() == consumerCount;
        });
    }

    public void untilConsumersStop() {
        sleep(3);
    }

    public void untilConsumersStart() {
        sleep(3);
    }

    public void untilConsumerCommitsOffset() {
        sleep(4);
    }

    public void untilPublishedMetricsPropagation() {
        sleep(2);
    }

    public void untilMessageDiscarded() {
        sleep(5);
    }

    public void untilMessageTraceLogged(final DBCollection collection, final PublishedMessageTraceStatus status) {
        waitAtMost(adjust(Duration.ONE_MINUTE)).until(() -> collection.find(new BasicDBObject("status", status.toString())).count() > 0);
    }

    public void untilMessageTraceLogged(final DBCollection collection, final SentMessageTraceStatus status) {
        waitAtMost(adjust(Duration.ONE_MINUTE)).until(() -> collection.find(new BasicDBObject("status", status.toString())).count() > 0);
    }

    public void untilMessageIdLogged(final DBCollection collection, final String messageId) {
        waitAtMost(adjust(Duration.ONE_MINUTE)).until(() -> collection.find(new BasicDBObject("messageId", messageId)).count() > 0);
    }

    public void untilReceivedAnyMessage(final DBCollection collection) {
        waitAtMost(adjust(Duration.ONE_MINUTE)).until(() -> collection.find().count() > 0);
    }

    private String subscriptionConsumerPath(String group, String topic, String subscription) {
        TopicName topicName = new TopicName(group, topic);
        return "/consumers/" + Subscription.getId(topicName, subscription) + "/owners/" + topicName.qualifiedName();
    }

    private String subscriptionIdsPath(String group, String topic, String subscription) {
        TopicName topicName = new TopicName(group, topic);
        return "/consumers/" + Subscription.getId(topicName, subscription) + "/ids";
    }

    private String subscriptionOffsetPath(String group, String topic, String subscription) {
        TopicName topicName = new TopicName(group, topic);
        return "/consumers/" + Subscription.getId(topicName, subscription) + "/offsets/" + topicName.qualifiedName();
    }

    private void sleep(int seconds) {
        try {
            Thread.sleep(adjust(seconds * 1000));
        } catch (InterruptedException exception) {
            throw new RuntimeException("Who dares to interrupt me?", exception);
        }
    }

    private void untilZookeeperNodeEmptied(final String path, int seconds, final CuratorFramework zookeeper) {
        waitAtMost(adjust(seconds), TimeUnit.SECONDS).until(() -> {
            List<String> children = zookeeper.getChildren().forPath(path);
            return children == null || children.isEmpty();
        });
    }

    private void untilZookeeperNodeCreation(final String path, final CuratorFramework zookeeper) {
        waitAtMost(adjust(20), TimeUnit.SECONDS).until(() -> zookeeper.checkExists().forPath(path) != null);
    }

    private void untilZookeeperNodeDeletion(final String path, final CuratorFramework zookeeper) {
        waitAtMost(adjust(Duration.FIVE_SECONDS)).until(() -> zookeeper.checkExists().forPath(path) == null);
    }

    public void waitUntilConsumerMetadataAvailable(Subscription subscription, String host, int port) {
        BlockingChannel channel = createBlockingChannel(host, port);
        channel.connect();

        waitAtMost(adjust((Duration.ONE_MINUTE))).until(() -> {
            channel.send(new ConsumerMetadataRequest(subscription.getId(), ConsumerMetadataRequest.CurrentVersion(), 0, "0"));
            ConsumerMetadataResponse metadataResponse = ConsumerMetadataResponse.readFrom(channel.receive().buffer());
            return metadataResponse.errorCode() == ErrorMapping.NoError();
        });

        channel.disconnect();
    }

    private BlockingChannel createBlockingChannel(String host, int port) {
        return new BlockingChannel(host, port,
                BlockingChannel.UseDefaultBufferSize(),
                BlockingChannel.UseDefaultBufferSize(),
                (int) adjust(Duration.TEN_SECONDS).getValueInMS());
    }
}
